#
# Copyright (c) 2016-2017 Balabit
#
# This file is part of Furnace.
#
# Furnace is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# Furnace is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Furnace.  If not, see <http://www.gnu.org/licenses/>.
#
import json
import logging
import os
import signal
import subprocess
import sys
from pathlib import Path
from typing import Union, List

from . import pid1
from .config import NAMESPACES, HOST_NETWORK_BIND_MOUNTS, BindMount
from .libc import unshare, setns, CLONE_NEWPID
from .utils import PathEncoder

logger = logging.getLogger(__name__)


class ContainerPID1Manager:
    def __init__(self, root_dir: Path, *, isolate_networking=False, bind_mounts=None):
        self.root_dir = root_dir.resolve()
        self.isolate_networking = isolate_networking
        self.bind_mounts = bind_mounts
        if self.bind_mounts is None:
            self.bind_mounts = []

    def do_exec(self, control_read, control_write):
        logger.debug("Executing {} {}".format(sys.executable, pid1.__file__))
        params = json.dumps({
            "loglevel": logging.getLevelName(logger.getEffectiveLevel()),
            "root_dir": self.root_dir,
            "control_read": control_read,
            "control_write": control_write,
            "isolate_networking": self.isolate_networking,
            "bind_mounts": self.bind_mounts,
        }, cls=PathEncoder)

        os.execl(sys.executable, sys.executable, pid1.__file__, params)

    def wait_for_ready_signal(self):
        if os.read(self.control_read, 3) != b"RDY":
            raise RuntimeError("Container PID 1 did not send Ready signal")

    def start(self):
        pipe_parent_read, pipe_child_write = os.pipe()
        pipe_child_read, pipe_parent_write = os.pipe()
        os.set_inheritable(pipe_child_read, True)
        os.set_inheritable(pipe_child_write, True)

        # We unshare (change) the pid namespace here, and other namespaces after
        # the exec, because if we exec'd in the new mount namespace, it would open
        # files in the new namespace's root, and prevent us from umounting the old
        # root after pivot_root. Note that changing the pid namespace affects only
        # the children (namely, which namespace they will be put in). It is thread
        # safe because unshare() affects the calling thread only.
        unshare(CLONE_NEWPID)

        self.pid = os.fork()
        if not self.pid:
            # this is the child process, will turn into PID1 in the container
            try:
                # this method will NOT return
                self.do_exec(pipe_child_read, pipe_child_write)
            except BaseException as e:
                # We are the child process, do NOT run parent's __exit__ handlers
                print(e, file=sys.stderr)
                os._exit(1)

        logger.debug("Container PID1 actual PID: {}".format(self.pid))

        # Reset the pid namespace of the parent process. /proc/self/ns/pid contains
        # a reference to the original pid namespace of the thread. New child processes
        # will be placed in this pid namespace after the setns() has restored the original
        # pid namespace
        original_pidns_fd = os.open('/proc/self/ns/pid', os.O_RDONLY)
        setns(original_pidns_fd, CLONE_NEWPID)
        os.close(original_pidns_fd)

        os.close(pipe_child_read)
        os.close(pipe_child_write)
        self.control_read = pipe_parent_read
        self.control_write = pipe_parent_write
        self.wait_for_ready_signal()

    def kill(self):
        # Killing pid1 will kill every other process in the context
        # The context itself will implode without any references,
        # basically cleaning up everything
        os.kill(self.pid, signal.SIGKILL)
        os.waitpid(self.pid, 0)


class SetnsContext:
    def __init__(self, pid):
        self.pid = pid
        # we open and close the ns file descriptors in the constructor
        # and 'destructor' for two reasons:
        # - if the context is used more than one time, it saves us the file opening
        #   neither the original ns, nor the container's ns is expected to change
        #   during the lifetime of this object
        # - we have to do it in a separate step from setns() calls, because after
        #   a mount namespace change, the next open might not work
        self.orig_fds = []
        self.new_fds = []
        self.orig_pidns = None
        self.new_pidns = None
        for ns_name, ns_flag in NAMESPACES.items():
            orig_ns_fd = os.open('/proc/self/ns/{}'.format(ns_name), os.O_RDONLY)
            new_ns_fd = os.open('/proc/{}/ns/{}'.format(self.pid, ns_name), os.O_RDONLY)
            if ns_flag == CLONE_NEWPID:
                self.orig_pidns = orig_ns_fd
                self.new_pidns = new_ns_fd
            else:
                self.orig_fds.append((orig_ns_fd, ns_flag))
                self.new_fds.append((new_ns_fd, ns_flag))

    def __del__(self):
        for fd, _ in self.orig_fds + self.new_fds:
            os.close(fd)
        os.close(self.orig_pidns)
        os.close(self.new_pidns)

    def __enter__(self):
        try:
            setns(self.new_pidns, CLONE_NEWPID)
        except Exception:
            self.__exit__(*sys.exc_info())
            raise
        return self

    def post_fork(self):
        for new_ns_fd, ns_flag in self.new_fds:
            setns(new_ns_fd, ns_flag)

    def __exit__(self, type, value, traceback):
        setns(self.orig_pidns, CLONE_NEWPID)
        return False


class ContainerContext:
    def __init__(self, root_dir: Union[str, Path], *, isolate_networking: bool = False, bind_mounts: List[BindMount] = None):
        if not isinstance(root_dir, Path):
            root_dir = Path(root_dir)
        self.root_dir = root_dir.resolve()
        if bind_mounts is None:
            bind_mounts = []
        if not isolate_networking:
            bind_mounts.extend(HOST_NETWORK_BIND_MOUNTS)
        self.pid1 = ContainerPID1Manager(root_dir, isolate_networking=isolate_networking, bind_mounts=bind_mounts)
        self.setns_context = None

    def __enter__(self):
        self.pid1.start()
        self.setns_context = SetnsContext(self.pid1.pid)
        return self

    def __exit__(self, type, value, traceback):
        self.setns_context = None
        self.pid1.kill()
        return False

    def run(self, *args, **kwargs):
        with self.setns_context:
            return subprocess.run(*args, **kwargs, preexec_fn=self.setns_context.post_fork)

    def Popen(self, *args, **kwargs):
        with self.setns_context:
            return subprocess.Popen(*args, **kwargs, preexec_fn=self.setns_context.post_fork)

    def interactive_shell(self, virtual_hostname='container'):
        print()
        self.run(
            ['bash', '--norc', '--noprofile', '-i'],
            env={
                'PS1': r'furnace-debug@{} \033[32m\w\033[0m # '.format(virtual_hostname)
            }
        )



#
# Copyright (c) 2016-2017 Balabit
#
# This file is part of Furnace.
#
# Furnace is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# Furnace is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Furnace.  If not, see <http://www.gnu.org/licenses/>.
#

import abc
import logging
from json import JSONEncoder
from pathlib import Path

from .libc import mount, umount, umount2, MS_BIND, MNT_DETACH, MS_REMOUNT, MS_RDONLY

logger = logging.getLogger(__name__)


class PathEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Path):
            return str(obj)
        return super().default(obj)


class MountContext(abc.ABC):
    def __init__(self, source, destination):
        self.source = source
        self.destination = destination

    def umount(self):
        logger.debug("Unmounting {path}".format(path=self.destination))
        try:
            umount(self.destination)
        except OSError:
            logger.warning("Failed to umount {path}, detaching instead".format(path=self.destination))
            umount2(self.destination, MNT_DETACH)

    @abc.abstractmethod
    def get_mount_parameters(self):
        pass

    def mount(self):
        fstype, flags, options = self.get_mount_parameters()
        mount(self.source, self.destination, fstype, flags, options)

    def __enter__(self):
        self.mount()
        return self

    def __exit__(self, type, value, traceback):
        self.umount()


class BindMountContext(MountContext):
    def __init__(self, source, destination, read_only=False):
        super().__init__(source, destination)
        self.read_only = read_only

    def get_mount_parameters(self):
        return None, MS_BIND, None

    def mount(self):
        super().mount()
        if self.read_only:
            mount(Path(), self.destination, None, MS_REMOUNT | MS_BIND | MS_RDONLY, None)


class OverlayfsMountContext(MountContext):
    def __init__(self, ro_dirs, rw_dir, work_dir, destination):
        super().__init__("overlay", destination)
        self.ro_dirs = ro_dirs
        self.rw_dir = rw_dir
        self.work_dir = work_dir

    def get_mount_parameters(self):
        options_string = 'lowerdir={lowerdir},upperdir={upperdir},workdir={workdir}'.format(
            lowerdir=':'.join([str(path) for path in self.ro_dirs]),
            upperdir=self.rw_dir,
            workdir=self.work_dir
        )
        return "overlay", 0, options_string



#
# Copyright (c) 2016-2017 Balabit
#
# This file is part of Furnace.
#
# Furnace is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# Furnace is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Furnace.  If not, see <http://www.gnu.org/licenses/>.
#

import json
import logging
import os
import signal
import stat
import subprocess
import sys
from socket import sethostname
from pathlib import Path

from furnace.libc import unshare, mount, umount2, non_caching_getpid, pivot_root, is_mount_point, \
    MS_BIND, MS_REC, MS_SLAVE, MS_REMOUNT, MS_RDONLY, CLONE_NEWPID, CLONE_NEWNET, MNT_DETACH
from furnace.config import NAMESPACES, CONTAINER_MOUNTS, CONTAINER_DEVICE_NODES, HOSTNAME, BindMount, DeviceNode

logger = logging.getLogger("container.pid1")


class PID1:
    def __init__(self, root_dir, control_read, control_write, isolate_networking, bind_mounts):
        self.control_read = control_read
        self.control_write = control_write
        self.root_dir = Path(root_dir).resolve()
        self.isolate_networking = isolate_networking
        self.bind_mounts = self.convert_bind_mounts_parameter(bind_mounts)
        self.loop_devices = list(self.get_loop_devices())

    @classmethod
    def convert_bind_mounts_parameter(cls, bind_mounts):
        result = []
        for source, destination, read_only in bind_mounts:
            source = Path(source)
            destination = Path(destination)
            if destination.is_absolute():
                destination = destination.relative_to("/")
            result.append(BindMount(source, destination, read_only))
        return result

    def enable_zombie_reaping(self):
        # We are pid 1, so we have to take care of orphaned processes
        # Interestingly, SIG_IGN is the default handler for SIGCHLD,
        # but this way we signal to the kernel that we will not call waitpid
        # and get rid of zombies automatically
        signal.signal(signal.SIGCHLD, signal.SIG_IGN)

    @classmethod
    def create_mount_target(cls, source, destination):
        if source.is_file():
            if destination.is_symlink():
                destination.unlink()
            destination.parent.mkdir(parents=True, exist_ok=True)
            destination.touch()
        else:
            destination.mkdir(parents=True, exist_ok=True)

    def create_bind_mounts(self):
        for source, relative_destination, read_only in self.bind_mounts:
            destination = self.root_dir.joinpath(relative_destination)
            self.create_mount_target(source, destination)
            mount(source, destination, None, MS_BIND, None)
            if read_only:
                # "Read-only bind mounts" are actually an illusion, a special feature of the kernel,
                # which is why we have to make the bind mount read-only in a separate call.
                # See https://lwn.net/Articles/281157/
                flags = MS_REMOUNT | MS_BIND | MS_RDONLY
                mount(Path(), destination, None, flags, None)

    def setup_root_mount(self):
        # SLAVE means that mount events will get inside the container, but
        # mounting something inside will not leak out.
        # Use PRIVATE to not let outside events propagate in
        mount(Path("none"), Path("/"), None, MS_REC | MS_SLAVE, None)
        self.create_bind_mounts()
        if not is_mount_point(self.root_dir):
            mount(self.root_dir, self.root_dir, None, MS_BIND, None)
        old_root_dir = self.root_dir.joinpath('old_root')
        old_root_dir.mkdir(parents=True, exist_ok=True)
        os.chdir(str(self.root_dir))
        pivot_root(Path('.'), Path('old_root'))
        os.chroot('.')

    def mount_defaults(self):
        for m in CONTAINER_MOUNTS:
            options = None
            if m.options:
                options = ",".join(m.options)
            m.destination.mkdir(parents=True, exist_ok=True)
            mount(m.source, m.destination, m.type, m.flags, options)

    def create_tmpfs_dirs(self):
        if Path('/bin/systemd-tmpfiles').exists():
            for m in CONTAINER_MOUNTS:
                if m.type == "tmpfs":
                    tmpfiles_output = subprocess.check_output(
                        ['/bin/systemd-tmpfiles', '--create', '--prefix', str(m.destination)],
                        stderr=subprocess.STDOUT,
                    )
                    if tmpfiles_output:
                        logger.debug("systemd-tmpfiles output: {}".format(tmpfiles_output))
        else:
            logger.warning(
                "Could not run systemd-tmpfiles, because it does not exist. "
                "/tmp and /run will not be populated."
            )

    def create_device_node(self, name, major, minor, mode, *, is_block_device=False):
        if is_block_device:
            device_type = stat.S_IFBLK
        else:
            device_type = stat.S_IFCHR
        nodepath = Path("/dev", name)
        os.mknod(str(nodepath), mode=device_type, device=os.makedev(major, minor))
        # A separate chmod is necessary, because mknod (undocumentedly) takes umask into account when creating
        nodepath.chmod(mode=mode)

    def create_default_dev_nodes(self):
        for d in CONTAINER_DEVICE_NODES:
            self.create_device_node(d.name, d.major, d.minor, 0o666)

    def create_loop_devices(self):
        self.create_device_node('loop-control', 10, 237, 0o660)
        for loop in self.loop_devices:
            self.create_device_node(loop.name, 7, loop.minor, 0o660, is_block_device=True)

    def umount_old_root(self):
        umount2('/old_root', MNT_DETACH)
        os.rmdir('/old_root')

    def create_namespaces(self):
        unshare_flags = 0
        for name, flag in NAMESPACES.items():
            if flag == CLONE_NEWPID:
                continue
            if flag == CLONE_NEWNET and not self.isolate_networking:
                continue
            if Path('/proc/self/ns', name).exists():
                unshare_flags = unshare_flags | flag
            else:
                logger.warning("Namespace type {} not supported on this system".format(name))
        unshare(unshare_flags)

    def run(self):
        if non_caching_getpid() != 1:
            raise ValueError("We are not actually PID1, exiting for safety reasons")

        # codecs are loaded dynamically, and won't work when we remount root
        make_sure_codecs_are_loaded = b'a'.decode('unicode_escape')  # NOQA: F841 local variable 'make_sure_codecs_are_loaded' is assigned to but never used
        os.setsid()
        self.enable_zombie_reaping()
        self.create_namespaces()
        self.setup_root_mount()
        self.mount_defaults()
        self.create_default_dev_nodes()
        self.create_loop_devices()
        self.create_tmpfs_dirs()
        self.umount_old_root()
        sethostname(HOSTNAME)

        os.write(self.control_write, b"RDY")
        logger.debug("Container started")
        # this will return when the pipe is closed
        # E.g. the outside control process died before killing us
        os.read(self.control_read, 1)
        logger.debug("Control pipe closed, stopping")
        return 0

    # NOTE: use only before create_namespaces()
    def get_loop_devices(self):
        for loop_path in Path('/dev').glob('loop[0-9]*'):
            major, minor = divmod(os.stat(str(loop_path)).st_rdev, 256)
            if major == 7:  # it's a loop device
                yield DeviceNode(name=loop_path.name, major=major, minor=minor)


if __name__ == "__main__":
    args = json.loads(sys.argv[1])
    logger.setLevel(args.pop("loglevel"))
    pid1 = PID1(**args)
    sys.exit(pid1.run())



#
# Copyright (c) 2016-2020 Balabit
#
# This file is part of Furnace.
#
# Furnace is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# Furnace is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Furnace.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import pytest
import re
import subprocess
import threading
from pathlib import Path

from furnace.config import BindMount
from furnace.context import ContainerContext
from furnace.libc import is_mount_point
from furnace.utils import BindMountContext, OverlayfsMountContext


@pytest.fixture(scope="session")
def debootstrapped_dir(tmp_path_factory):
    result = os.environ.get("DEBOOTSTRAPPED_DIR")
    if not result:
        result = str(tmp_path_factory.mktemp('debootstrapped_dir'))
        subprocess.run(['debootstrap', 'xenial', result, 'http://archive.ubuntu.com/ubuntu'], check=True)
    yield Path(result)


@pytest.fixture
def rootfs_for_testing(debootstrapped_dir, tmp_path):
    overlay_workdir = tmp_path.joinpath('overlay_work')
    overlay_workdir.mkdir()
    overlay_rwdir = tmp_path.joinpath('overlay_rw')
    overlay_rwdir.mkdir()
    overlay_mounted = tmp_path.joinpath('overlay_mount')
    overlay_mounted.mkdir()
    with OverlayfsMountContext([debootstrapped_dir], overlay_rwdir, overlay_workdir, overlay_mounted):
        yield overlay_mounted


def test_container_basic(rootfs_for_testing):
    cwd = os.getcwd()
    with ContainerContext(rootfs_for_testing) as cnt:
        ps_output = cnt.run(['/bin/ps', '-e', '-o', 'pid,command', '--no-headers'], check=True, stdout=subprocess.PIPE).stdout
        ps_output = ps_output.decode('utf-8').split('\n')
        assert cwd == os.getcwd(), "Container.run() should not touch CWD"
        assert len(ps_output) > 0, 'There should be at least one process in the container (the ps). Might "ps" have failed?'
        assert len(ps_output) <= 3, \
            'There should be at most 3 processes in the container ' \
            '(init, ps, and possibly bash that runs ps).' \
            'Is it even contained? Is the proper /proc mounted?'

        false_output = cnt.run(['/bin/sh', '-c', 'exit 42'])
        assert false_output.returncode == 42, "Return codes of command should be preserved"

    assert cwd == os.getcwd(), "Container should not touch CWD"


def test_container_rootfs_is_not_mountpoint(debootstrapped_dir):
    assert not is_mount_point(debootstrapped_dir)
    with ContainerContext(debootstrapped_dir):
        pass


def test_container_rootfs_is_mountpoint(rootfs_for_testing):
    assert is_mount_point(rootfs_for_testing)
    with ContainerContext(rootfs_for_testing):
        pass


def test_mounts_visible_inside_container(rootfs_for_testing, tmp_path):
    with ContainerContext(rootfs_for_testing) as cnt:
        tmp_path.joinpath('test_file').write_text('Test data')
        mounted_after_path = rootfs_for_testing.joinpath('mounted_after')
        mounted_after_path.mkdir()
        with BindMountContext(tmp_path, mounted_after_path):
            output = cnt.run(['/bin/cat', '/mounted_after/test_file'], check=True, stdout=subprocess.PIPE).stdout
            assert output == b"Test data"


def test_files_made_in_container_visible_outside(rootfs_for_testing):
    with ContainerContext(rootfs_for_testing) as cnt:
        cnt.run(['/bin/dd', 'if=/dev/zero', 'of=/test_file', 'bs=128', 'count=1'], check=True)
        with rootfs_for_testing.joinpath('test_file').open('rb') as f:
            file_data = f.read()
        assert file_data == b'\0' * 128


def test_lingering_processes_are_killed(rootfs_for_testing):
    with ContainerContext(rootfs_for_testing) as cnt:
        cnt.run(['/bin/bash', '-c', '/bin/sleep 31337 >/dev/null 2>/dev/null&'], check=True)
        ps_output = subprocess.run(['/bin/ps', 'aux'], check=True, stdout=subprocess.PIPE).stdout.decode('utf-8')
        assert 'sleep 31337' in ps_output, 'The sleep should be running in the container'
    ps_output = subprocess.run(['/bin/ps', 'aux'], check=True, stdout=subprocess.PIPE).stdout.decode('utf-8')
    assert 'sleep 31337' not in ps_output, 'The sleep should no longer be running'


def test_zombies_are_killed(rootfs_for_testing):
    with ContainerContext(rootfs_for_testing) as cnt:
        # the || true is needed, otehrwise bash will exec the kill
        try:
            cnt.run(['/bin/bash', '-c', '/bin/bash -c "kill -9 $$ || true" || true'], check=True)
        except subprocess.CalledProcessError:
            pass
        ps_output = cnt.run(['/bin/ps', 'aux'], check=True, stdout=subprocess.PIPE).stdout.decode('utf-8')
        assert '<defunct>' not in ps_output, ''


def test_lock_dirs_are_present(rootfs_for_testing):
    with ContainerContext(rootfs_for_testing) as cnt:
        cnt.run(['/usr/bin/test', '-e', '/var/lock'], check=True)
        cnt.run(['/usr/bin/test', '-e', '/run/lock'], check=True)
        # no assert, because the previous two commands would have thrown an Exception on error


def test_networking_is_isolated_when_asked(rootfs_for_testing):
    with ContainerContext(rootfs_for_testing, isolate_networking=True) as cnt:
        ip_output = cnt.run(['/bin/ip', 'address', 'list'], check=True, stdout=subprocess.PIPE).stdout.decode('utf-8')
        assert re.search("^1: lo", ip_output, flags=re.MULTILINE) is not None, "Loopback interface should be present"
        assert re.search("^2: ", ip_output, flags=re.MULTILINE) is None, "No other interfaces should be present"


def test_networking_is_not_isolated_when_asked(rootfs_for_testing):
    with ContainerContext(rootfs_for_testing, isolate_networking=False) as cnt:
        ip_output = cnt.run(['/bin/ip', 'address', 'list'], check=True, stdout=subprocess.PIPE).stdout.decode('utf-8')
        assert re.search("^1: lo", ip_output, flags=re.MULTILINE) is not None, "Loopback interface should be present"
        assert re.search("^2: ", ip_output, flags=re.MULTILINE) is not None, "At least one other interface should be present"


def test_loop_mounts_work(rootfs_for_testing):
    with ContainerContext(rootfs_for_testing) as cnt:
        cnt.run(['/bin/dd', 'if=/dev/zero', 'of=/disk.img', 'bs=1M', 'count=10'], check=True)
        cnt.run(['/sbin/mkfs.ext4', '/disk.img'], check=True)
        rootfs_for_testing.joinpath('mounted').mkdir()
        cnt.run(['/bin/mount', '-o', 'loop', '/disk.img', '/mounted'], check=True)
        cnt.run(['/bin/touch', '/mounted/test.file'], check=True)
        # no assert, because the previous two commands would have thrown an Exception on error


def test_using_container_does_not_touch_files_if_network_isolated(debootstrapped_dir, tmp_path):
    overlay_workdir = tmp_path.joinpath('overlay_work')
    overlay_workdir.mkdir()
    overlay_rwdir = tmp_path.joinpath('overlay_rw')
    overlay_rwdir.mkdir()
    overlay_mounted = tmp_path.joinpath('overlay_mount')
    overlay_mounted.mkdir()
    with OverlayfsMountContext([debootstrapped_dir], overlay_rwdir, overlay_workdir, overlay_mounted):
        with ContainerContext(overlay_mounted, isolate_networking=True) as cnt:
            cnt.run(["/bin/ls", "/"], check=True)

    modified_files = list(overlay_rwdir.iterdir())
    assert len(modified_files) == 0, "No files should have been modified because of container run if network isolated"


def test_using_container_with_host_network(rootfs_for_testing, tmp_path):
    host_resolvconf_content = Path('/etc/resolv.conf').read_bytes()
    container_resolvconf_path = rootfs_for_testing.joinpath('etc', 'resolv.conf')
    if container_resolvconf_path.exists():
        container_resolvconf_path.unlink()
    with ContainerContext(rootfs_for_testing) as cnt:
        cnt.run(["/bin/ls", '-la', "/"], check=True)
        cnt.run(["/bin/mount"], check=True)

        touch_result = cnt.run(["/bin/touch", "/etc/resolv.conf"])
        assert touch_result.returncode != 0, "Touch should fail on resolv.conf, as it should be read-only"

        container_resolvconf_content = cnt.run(["/bin/cat", "/etc/resolv.conf"], check=True, stdout=subprocess.PIPE).stdout

    assert host_resolvconf_content == container_resolvconf_content, \
        "The content of '/etc/resolv.conf' of the host machine should be equal to '/etc/resolv.conf' of the container " \
        "if host networking is used"


class ThreadForTesting(threading.Thread):
    def __init__(self):
        super().__init__(name='container-test-ing-thread', daemon=True)
        self.stopme = False

    def run(self):
        # This thread is intentionally busy to test for
        # context-switch race conditions
        lcg = 1
        while self.stopme:
            lcg = (lcg * 1234567 + 1) % 87654321

    def stop(self):
        self.stopme = True
        self.join()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *args, **kwargs):
        self.stop()


def test_thread_support(rootfs_for_testing):
    with ThreadForTesting():
        with ContainerContext(rootfs_for_testing) as cnt:
            cnt.run(['/bin/echo', 'Hello world'])
            cnt.Popen(['/bin/echo', 'Hello Popen']).wait()


def test_bind_mounts(rootfs_for_testing, tmp_path):
    tmp_path.joinpath('test_file').write_text('Test data')
    bind_mounts = [
        BindMount(tmp_path, Path('mounted', 'multiple', 'dirs'), False),
        BindMount(tmp_path, Path('mounted_ro'), True),
        BindMount(tmp_path.joinpath('test_file'), Path('mounted', 'as', 'file'), False),
        BindMount(tmp_path, Path('/', 'mounted', 'as', 'absolute'), False),
    ]

    with ContainerContext(rootfs_for_testing, bind_mounts=bind_mounts) as cnt:
        output = cnt.run(['/bin/cat', '/mounted/multiple/dirs/test_file'], check=True, stdout=subprocess.PIPE).stdout
        assert output == b"Test data"
        output = cnt.run(['/bin/cat', '/mounted_ro/test_file'], check=True, stdout=subprocess.PIPE).stdout
        assert output == b"Test data"
        output = cnt.run(['/bin/cat', '/mounted/as/file'], check=True, stdout=subprocess.PIPE).stdout
        assert output == b"Test data"
        output = cnt.run(['/bin/cat', '/mounted/as/absolute/test_file'], check=True, stdout=subprocess.PIPE).stdout
        assert output == b"Test data"
        result = cnt.run(['/bin/touch', '/mounted_ro/test_file'])
        assert result.returncode != 0, "Touch should fail, because mounted_ro should be read-only"

        assert not rootfs_for_testing.joinpath('mounted_ro', 'test_file').exists(), \
            "Bind mounts done by ContainerContext should not be visible outside of the container"


def test_ro_bind_mounts_from_outside(rootfs_for_testing, tmp_path):
    tmp_path.joinpath('test_file').write_text('Test data')

    mounted_before_path = rootfs_for_testing.joinpath('mounted_ro')
    mounted_before_path.mkdir()
    with BindMountContext(tmp_path, mounted_before_path, read_only=True):
        with pytest.raises(OSError):
            rootfs_for_testing.joinpath('mounted_ro', 'test_file').touch()
            pytest.fail("mounted_ro should be mounted readonly")
        with ContainerContext(rootfs_for_testing) as cnt:
            output = cnt.run(['/bin/cat', '/mounted_ro/test_file'], check=True, stdout=subprocess.PIPE).stdout
            assert output == b"Test data"
            result = cnt.run(['/bin/touch', '/mounted_ro/test_file'])
            assert result.returncode != 0, "Touch should fail, because mounted_ro should be read-only"



#
# Copyright (c) 2016-2017 Balabit
#
# This file is part of Furnace.
#
# Furnace is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# Furnace is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Furnace.  If not, see <http://www.gnu.org/licenses/>.
#

from .context import ContainerContext  # NOQA: F401 '.context.ContainerContext' imported but unused



#
# Copyright (c) 2016-2017 Balabit
#
# This file is part of Furnace.
#
# Furnace is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# Furnace is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Furnace.  If not, see <http://www.gnu.org/licenses/>.
#

import ctypes
import logging
from pathlib import Path

logger = logging.getLogger(__name__)

libc = ctypes.CDLL("libc.so.6", use_errno=True)


MS_RDONLY = 0x00000001
MS_NOSUID = 0x00000002
MS_NODEV = 0x00000004
MS_NOEXEC = 0x00000008
MS_REMOUNT = 0x00000020
MS_NOATIME = 0x00000400
MS_NODIRATIME = 0x00000800
MS_BIND = 0x00001000
MS_MOVE = 0x00002000
MS_REC = 0x00004000
MS_PRIVATE = 0x00040000
MS_SLAVE = 0x00080000
MS_SHARED = 0x00100000
MS_STRICTATIME = 0x01000000

CLONE_NEWNS = 0x00020000
CLONE_NEWCGROUP = 0x02000000
CLONE_NEWUTS = 0x04000000
CLONE_NEWIPC = 0x08000000
CLONE_NEWPID = 0x20000000
CLONE_NEWNET = 0x40000000

SYSCALL_NUM_CLONE = 56
SYSCALL_NUM_GETPID = 39

MNT_DETACH = 2


def mount(source: Path, target: Path, fstype, flags, data):
    if fstype is not None:
        fstype = fstype.encode('utf-8')
    if data is not None:
        data = data.encode('utf-8')
    if libc.mount(str(source).encode('utf-8'), str(target).encode('utf-8'), fstype, flags, data) != 0:
        raise OSError(ctypes.get_errno(), "Mount failed")


def umount(target: Path):
    return umount2(target, 0)


def umount2(target: Path, flags: int):
    if libc.umount2(str(target).encode('utf-8'), flags) != 0:
        raise OSError(ctypes.get_errno(), "Failed to unmount directory: {}; flags={}".format(target, flags))


def get_all_mounts():
    with Path("/proc/self/mounts").open('rb') as f:
        mounts = []
        for line in f:
            mp = line.split(b' ')[1]
            # unicode_escape is necessary because moutns with special characters are
            # represented with octal escape codes in 'mounts'
            mounts.append(Path(mp.decode('unicode_escape')))
    return mounts


def is_mount_point(path: Path):
    # os.path.ismount does not properly detect bind mounts
    return path.absolute() in get_all_mounts()


def unshare(flags):
    if libc.unshare(flags) != 0:
        raise OSError(ctypes.get_errno(), "unshare failed")


def setns(fd, flags):
    if libc.setns(fd, flags) != 0:
        raise OSError(ctypes.get_errno(), "setns failed")


def pivot_root(new_root: Path, old_root: Path):
    if libc.pivot_root(str(new_root).encode('utf-8'), str(old_root).encode('utf-8')) != 0:
        raise OSError(ctypes.get_errno(), "pivot_root failed")


def clone(flags, stack=0):
    syscall = libc.syscall
    syscall.restype = ctypes.c_int
    syscall.argtypes = (ctypes.c_int, ctypes.c_int, ctypes.c_int)
    result = syscall(SYSCALL_NUM_CLONE, flags, stack)
    if result < 0:
        raise OSError(abs(result), "clone failed")
    return result


def non_caching_getpid():
    # libc caches the return value of getpid, and does not refresh this
    # cache, if we call syscalls (e.g. clone) by hand.
    syscall = libc.syscall
    syscall.restype = ctypes.c_int
    syscall.argtypes = (ctypes.c_int,)
    result = syscall(SYSCALL_NUM_GETPID)
    if result < 0:
        raise OSError(abs(result), "getpid failed")
    return result



#!/usr/bin/env python3
# Copyright (c) 2016-2020 Balabit
#
# This file is part of Furnace.
#
# Furnace is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# Furnace is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Furnace.  If not, see <http://www.gnu.org/licenses/>.
#
import os
import sys

from pathlib import Path
from setuptools import setup, find_packages
from furnace.version import get_version


if sys.version_info < (3, 6):
    sys.exit('Python version 3.6+ is required for furnace')

with open(os.path.join(os.path.dirname(__file__), 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='furnace',
    version=get_version(),
    description='A lightweight pure-python container implementation',
    long_description=long_description,
    url='https://github.com/balabit/furnace',
    keywords='containers containerization',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Topic :: System',
    ],
    packages=find_packages(include=["furnace*"]),
    extras_require={
        'dev': Path('requirements-dev.txt').read_text()
    },
    package_data={'furnace': [
        'VERSION',
    ]},
    python_requires=">=3.6",
)



#
# Copyright (c) 2016-2017 Balabit
#
# This file is part of Furnace.
#
# Furnace is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# Furnace is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Furnace.  If not, see <http://www.gnu.org/licenses/>.
#

from collections import namedtuple

from pathlib import Path

from .libc import CLONE_NEWPID, CLONE_NEWCGROUP, CLONE_NEWIPC, CLONE_NEWUTS, CLONE_NEWNS, CLONE_NEWNET, \
    MS_NOSUID, MS_NOEXEC, MS_NODEV, MS_RDONLY, MS_STRICTATIME

Mount = namedtuple('Mount', ['destination', 'type', 'source', 'flags', 'options'])
DeviceNode = namedtuple('DeviceNode', ['name', 'major', 'minor'])
BindMount = namedtuple('BindMount', ['source', 'destination', 'readonly'])


HOSTNAME = 'localhost'

NAMESPACES = {
    "pid": CLONE_NEWPID,
    "cgroup": CLONE_NEWCGROUP,
    "ipc": CLONE_NEWIPC,
    "uts": CLONE_NEWUTS,
    "mnt": CLONE_NEWNS,
    "net": CLONE_NEWNET,
}

CONTAINER_MOUNTS = [
    Mount(
        destination=Path("/proc"),
        type="proc",
        source="proc",
        flags=0,
        options=None,
    ),
    Mount(
        destination=Path("/dev"),
        type="tmpfs",
        source="tmpfs",
        flags=MS_NOSUID | MS_STRICTATIME,
        options=[
            "mode=755",
            "size=65536k",
        ],
    ),
    Mount(
        destination=Path("/dev/pts"),
        type="devpts",
        source="devpts",
        flags=MS_NOSUID | MS_NOEXEC,
        options=[
            "newinstance",
            "ptmxmode=0666",
            "mode=0620",
            "gid=5",
        ],
    ),
    Mount(
        destination=Path("/dev/shm"),
        type="tmpfs",
        source="shm",
        flags=MS_NOSUID | MS_NOEXEC | MS_NODEV,
        options=[
            "mode=1777",
            "size=65536k",
        ],
    ),
    Mount(
        destination=Path("/dev/mqueue"),
        type="mqueue",
        source="mqueue",
        flags=MS_NOSUID | MS_NOEXEC | MS_NODEV,
        options=None,
    ),
    Mount(
        destination=Path("/sys"),
        type="sysfs",
        source="sysfs",
        flags=MS_NOSUID | MS_NOEXEC | MS_NODEV | MS_RDONLY,
        options=None,
    ),
    Mount(
        destination=Path("/run"),
        type="tmpfs",
        source="shm",
        flags=MS_NOSUID | MS_NOEXEC | MS_NODEV,
        options=[
            "mode=1777",
            "size=65536k",
        ],
    ),
]

CONTAINER_DEVICE_NODES = [
    DeviceNode(
        name="null",
        major=1,
        minor=3,
    ),
    DeviceNode(
        name="zero",
        major=1,
        minor=5,
    ),
    DeviceNode(
        name="full",
        major=1,
        minor=7,
    ),
    DeviceNode(
        name="tty",
        major=5,
        minor=0,
    ),
    DeviceNode(
        name="random",
        major=1,
        minor=8,
    ),
    DeviceNode(
        name="urandom",
        major=1,
        minor=9,
    ),
]

HOST_NETWORK_BIND_MOUNTS = [
    BindMount(
        source=Path('/etc/resolv.conf'),          # path on host machine
        destination=Path('/etc/resolv.conf'),     # path in container
        readonly=True,
    ),
]
